#include "mex.h"
#include "timer.h"
#include "main_algo.h"

// *****************************************************************************
// mexFunction
// *****************************************************************************

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	/* correct parameters:
		0) scalar number N of variables (variables must be 1:N).
		1) scalar number M of factors
    2) 1xN vector with the variables' domain sizes
		3) full Mx1 cell array with factor structs.
		   Each factor F has the following fields:
		   F.scope: 1xD scope of the factor, i.e. variables V over which it is defined (all v in V must be between 1 and N)
 		   F.sizes: 1xD size of the variables in F.scope
		   F.T: the factor's table with dimensions d_scope[1] x d_scope[2] x ... x d_scope[D]. 
            The first variable in F.scope toggles fastest, and the last one slowest.
		4) Scalar time allocated to the search
		5) Integer representing the algorithm to run
		6) debug level (<= 0: no output. 100: all output)
	*/		   

	if (nrhs < 4) mexErrMsgTxt("Must provide at least 4 inputs: <N, M, sizes, factors> Optional 5th: time. Optional 6th: algo (1=varelim, 2=ILS, 3=GLS+)");
//	if (nlhs != 5) mexErrMsgTxt("Must be exactly 5 outputs.");
	if (nrhs > 7) mexErrMsgTxt("Too many input arguments!");

	int debug = 0;
	if (nrhs == 7) debug = mxGetScalar(prhs[6]);;

	int i, var, val, pot;

	if(debug>10) printf("Start initializing C factor graph\n");
	start_timer();

	//=== Get number of variables.
	num_vars = mxGetScalar(prhs[0]);

	//=== Get number of factors.
	num_pots = mxGetScalar(prhs[1]);

	if(debug>10) printf("... num_vars=%d, num_pots=%d\n", num_vars, num_pots);
	allocateVarsAndPTs(false);
	if(debug>10) printf("... Allocated vars and factors\n");

	//=== Get domain sizes.
	double* domSizes = mxGetPr(prhs[2]);
	for(var=0; var<num_vars; var++){
		int domSize = (int) floor(domSizes[var]+0.5);
//		printf("domSize[%d]=%d\n",var,domSize);
		mxAssertS(domSize > 0 , "All domain sizes must be greater than 0.");		
		variables[var]->setDomainSize(domSize);
	}

	//=== Get the factors.
	for(pot=0; pot<num_pots; pot++){
		if(debug>10) printf("... Factor #%d\n", pot);
		mxArray *factor = mxGetCell(prhs[3], pot);
		
		if(debug>20) printf("... Reading it's scope and T\n");
	//=== Read data of the factor struct.
		const mxArray *scope = mxGetField(factor, 0, "scope");
//		const mxArray *sizes = mxGetField(factor, 0, "sizes");
		const mxArray *T = mxGetField(factor, 0, "T");

		//=== Assert that the scope has the right number of dimensions.
		int numScopeDim = mxGetNumberOfDimensions(scope);
		mxAssertS(numScopeDim == 2, "factor.scope must have 2 dimensions. It has to be a 1xD vector)");		
		const int* scopeDims = mxGetDimensions(scope);
		mxAssertS(scopeDims[0] == 1, "factor.scope must be a 1xD vector)");		
		
		//=== Retrieve the number of variables of the factor.
		int numVarsInFactor = scopeDims[1];
		if(debug>20) printf("numVarsInFactor=%d\n",numVarsInFactor);

		//=== Initialize the new factor.
		probTables[pot]->init(numVarsInFactor);

		//=== Retrieve the variables of the factor and put them in.
		double* varsInFactor = mxGetPr(scope);
		int numEntries = 1;
		for(i=0; i<numVarsInFactor; i++){
			var = (int) floor(varsInFactor[i]+0.5)-1;
			if(debug>20) printf("%dth var=%d\n",i,var);
			mxAssertS(0<=var && var<num_vars, "all variables in original factor.scope must be in between 1 and N");		
			probTables[pot]->setVar((numVarsInFactor-1)-i, var);
			numEntries *= variables[var]->domSize;
		}

		if(debug>20) printf("... Putting in T's entries\n");
		//=== Set the number of entries.
		probTables[pot]->setNumEntries(numEntries);

		//=== Assert that the table T has the right number of dimensions.
		int numDimsT = mxGetNumberOfDimensions(T);
		if(debug>30) printf("numDimsT=%d, numVarsInFactor=%d\n", numDimsT, numVarsInFactor);
		mxAssertS(numDimsT == numVarsInFactor || (numDimsT == 2 && numVarsInFactor==1), "factor.T must be a K-dimensional matrix where K is the number of variables in the factor. (Its dimensions must equal the sizes of its variables)");

			//=== Assert that the dimensions of table T have the right size.
		const int* dimsOfTable = mxGetDimensions(T);
		for(i=0; i<numVarsInFactor; i++){
			if(debug>30) printf("dim %d: %d\n", i, dimsOfTable[i]);
			mxAssertS(dimsOfTable[i] == variables[(int) floor(varsInFactor[i]+0.5)-1]->domSize, 
			  "factor.T's dimensions must equal the sizes of its variables");
		}
		
		//=== Put in all the entries of table T.
		double* entries = mxGetPr(T);
		for(i=0; i<numEntries; i++){
			if(debug>50) printf("... Entry %d/%d\n",i,numEntries);
			probTables[pot]->setEntry(i,entries[i]); 
		}		
	}


	//=== Variable SLS parameters.
//	assignmentManager.M_MPE = true;
//	assignmentManager.M = (int) mxGetScalar(prhs[3]);
	
	maxTime = 1;
	algo = ALGO_ILS;
	if( nrhs > 4 ) maxTime = (double) mxGetScalar(prhs[4]);
	if( nrhs > 5 ) algo = mxGetScalar(prhs[5]);

	if(debug>5) printf("maxTime=%lf, algo=%d\n",maxTime, algo);
	
	//=== SLS parameters that are currently fixed, but just since I don't want to deal with too many parameters.
	init_algo = INIT_RANDOM; 
//	outputBestMPE = false;
	pertubationFixVars = true;
	caching = CACHING_GOOD_PQ; //CACHING_GOOD_VARS;
	assignmentManager.optimalLogMPEValue = DOUBLE_BIG; // no uppoer bound.

	//=== FIXED SLS parameters.
	preprocessingSizeBound = 0;  // I have *not* implemented the recovery of eliminated variables!! 
	                             // (this is straightforward but may be tedious and I was lazy)
	maxRuns = 1; // We're not interested in analyzing the algorithm, but in an actual solution to the problem.


	//=== RUN SLS algorithm to solve the MPE problem.
	double tranlateTime = elapsed_seconds();
	if(debug>5) printf("Reading stuff into C representation took %lf seconds\n", tranlateTime);
	if(debug>5) printf("Now calling SLS\n");
	int* result = new int[num_vars];
	double logLikelihood;
	runAlgorithm(&result, &logLikelihood);
	if(debug>5) printf("Done running SLS - it ran for %lf seconds\n", run_time_so_far);

	//=== Return the result.
	plhs[0] = mxCreateScalarDouble(assignmentManager.overallBestLogProb);

	int result1_dim[2] = {1,num_vars};
  plhs[1] = mxCreateNumericArray(2, result1_dim, mxDOUBLE_CLASS, mxREAL);
	double* res1Ptr = mxGetPr(plhs[1]);
  for (var=0; var<num_vars; var++) {
		res1Ptr[var] = result[var];
	}

	int provedOptimality = 0;
	if(assignmentManager.optimalLogMPEValue < assignmentManager.overallBestLogProb + EPS){
		provedOptimality=1;
	}
	plhs[2] = mxCreateScalarDouble(provedOptimality);
	plhs[3] = mxCreateScalarDouble(run_time_so_far+tranlateTime);

	//=== If we are returning more than 1 good solution, fill 2 more output arguments.
	if( assignmentManager.M_MPE ){
		int result3_dim[2] = {assignmentManager.M, 1};
	  plhs[4] = mxCreateNumericArray(2, result3_dim, mxDOUBLE_CLASS, mxREAL);
		double* res3Ptr = mxGetPr(plhs[3]);
	    for(i=0; i<assignmentManager.M; i++){
			res3Ptr[i] = assignmentManager.runBestLogProbs[i];
		}
	
		int result4_dim[2] = {assignmentManager.M, num_vars};
	  plhs[5] = mxCreateNumericArray(2, result4_dim, mxDOUBLE_CLASS, mxREAL);
		double* res4Ptr = mxGetPr(plhs[4]);
	    for(i=0; i<assignmentManager.M; i++){
			for (var=0; var<num_vars; var++) {
				res4Ptr[i + var*assignmentManager.M] = assignmentManager.runBestAssignments[i][var];
			}
		}
	}
	
	if(debug>5) printf("Done with C code\n");
}
/*
  // **************************************************************************
  // free memory
  // **************************************************************************

  delete algorithm;
  algorithm = 0;
  
  if (singleBeliefs != 0) {
    for (int i=0; i<num_nodes; i++) {
      delete[] singleBeliefs[i];
    }
    delete[] singleBeliefs;    
    singleBeliefs = 0;
  }
  if (pairBeliefs != 0 && algo_type == AT_GBP) {
    for (int i=0; i<num_nodes; i++) {
      for (int n=0; n<mrf->neighbNum(i); n++) {
	if (pairBeliefs[i][n] != 0) {
	  for (int xi=0; xi<mrf->V[i]; xi++) {
	    delete[] pairBeliefs[i][n][xi];
	  }
	  delete[] pairBeliefs[i][n];
	}
      }
      delete[] pairBeliefs[i];
    }
    delete[] pairBeliefs;
    pairBeliefs = 0;
  }
  if (processor != 0) {
    delete processor;
    processor = 0;
  }

  delete mrf;
  mrf = 0;
  delete adjMat;
  adjMat = 0;
  
}
*/